from mobo.cluster import BaseClusterer
from mobo.error import BaseErrorCalculator
from mobo.filter import BaseFilter
from mobo.log import Logger
from mobo.parameter import Parameter
from mobo.projection import BaseProjector
from mobo.qoi import QoI
from typing import List, Optional


class LocalConfiguration(object):
    """Container for information required in each iteration.
    
    Args:
        n_samples: Number of samples to draw.
        clusterer: Clustering algorithm.
        error_calculator: Error calculation scheme.
        filters: Filters to apply.
        projector: Dimensionality reduction scheme.
    """
    def __init__(self,
                 n_samples: int,
                 clusterer: BaseClusterer,
                 error_calculator: BaseErrorCalculator,
                 filters: List[BaseFilter],
                 projector: BaseProjector) -> None:
        self.n_samples = n_samples
        self.clusterer = clusterer
        self.error_calculator = error_calculator
        self.filters = filters
        self.projector = projector


class GlobalConfiguration(object):
    """Container for information required to start an optimization.
    
    Args:
        n_samples: Number of samples to start with.
        local_configurations: Configuration for each iteration.
        parameters: Parameter objects to fit.
        qois: QoI objects to evaluate.
        initial_data_path: Path to a data file to start from.
        logger: Logging utility to monitor progress of the optimization.
    """
    def __init__(self,
                 n_samples: int,
                 local_configurations: List[LocalConfiguration],
                 parameters: List[Parameter], 
                 qois: List[QoI], 
                 initial_data_path: Optional[str] = None,
                 logger: Optional[Logger] = None) -> None:
        self.n_samples = n_samples
        self.local_configurations = local_configurations
        self.parameters = parameters
        self.qois = qois
        self.initial_data_path = initial_data_path
        self.logger = logger



from abc import ABC
import numpy as np
from scipy.stats import zscore
from sklearn.preprocessing import normalize


class BaseFilter(ABC):
    """Abstract base class for Filters."""
    def __call__(self, data: np.ndarray) -> np.ndarray:
        pass


class ParetoFilter(BaseFilter):
    """Pareto optimality filter."""
    def __call__(self, data: np.ndarray) -> np.ndarray:
        mask = np.ones(data.shape[0], dtype=bool)
        for i, err in enumerate(data):
            if mask[i]:
                # keep points with a lower error
                mask[mask] = np.any(data[mask] < err, axis=1)
                # and keep self
                mask[i] = True
        return mask


class PercentileFilter(BaseFilter):
    """Percentile scoring filter."""
    def __init__(self, percentile: int = 95) -> None:
        self._percentile = percentile
    
    def __call__(self, data: np.ndarray) -> np.ndarray:
        normalized = np.absolute(normalize(data, axis=0))
        scores = np.sum(normalized, axis=1) # sum each row
        critical_score = np.percentile(scores, self._percentile)
        return np.array([s <= critical_score for s in scores])


class ZscoreFilter(BaseFilter):
    """Z-Score filter."""
    def __init__(self, z: float = -1.5) -> None:
        self._z = z

    def __call__(self, data: np.ndarray) -> np.ndarray:
        normalized = np.absolute(normalize(data, axis=0))
        scores = np.sum(normalized, axis=1) # sum each row
        z_values = zscore(scores)
        return np.array([z >= self._z for z in z_values])



from datetime import datetime


class Logger(object):
    """A centralized log file.

    Args:
        path: Path to the log file.
    """
    def __init__(self, path: str = "mobo.log") -> None:
        self.path = path

    def log(self, msg: str) -> None:
        """Writes a message to the log file.
        
        Args:
            msg: The message to write.
        """
        now = datetime.now()
        msg = "{}\n{}\n\n".format(now, msg)
        with open(self.path, "a") as f:
            f.write(msg)



from mobo.filter import ParetoFilter, PercentileFilter, ZscoreFilter
import numpy as np

DATA = np.random.normal(size=(1000, 3))


def test_pareto_filter():
    pareto = ParetoFilter()
    mask = pareto(DATA)
    filtered_data = DATA[mask]
    assert filtered_data.shape[0] < DATA.shape[0]

def test_percentile_filter():
    perc = PercentileFilter() # 95th percentile default
    mask = perc(DATA)
    filtered_data = DATA[mask]
    assert filtered_data.shape[0] == 950

def test_zscore_filter():
    zscore = ZscoreFilter() # -1.5 z score default
    mask = zscore(DATA)
    filtered_data = DATA[mask]
    assert filtered_data.shape[0] < DATA.shape[0]



from abc import ABC
import numpy as np
from sklearn.decomposition import PCA
from sklearn.manifold import MDS, TSNE
from typing import Callable, Optional, Union


class BaseProjector(ABC):
    """Abstract base class for Projectors."""
    def __call__(self, data: np.ndarray) -> np.ndarray: 
        pass


class MDSProjector(BaseProjector):
    """Multi-Dimensional Scaling projector.
    
    Args:
        Reference:
        https://scikit-learn.org/stable/modules/generated/sklearn.manifold.MDS.html#sklearn.manifold.MDS
    """
    def __init__(self,
                 metric: bool = True,
                 n_init: int = 4,
                 max_iter: int = 300,
                 verbose: int = 0,
                 eps: float = 1e-3,
                 random_state: Optional[int] = None,
                 dissimilarity: str = "euclidean") -> None:
        self._projector = MDS(n_components=2,
                              n_jobs=None,
                              metric=metric,
                              n_init=n_init,
                              max_iter=max_iter,
                              verbose=verbose,
                              eps=eps,
                              random_state=random_state,
                              dissimilarity=dissimilarity)

    def __call__(self, data: np.ndarray) -> np.ndarray:
        return self._projector.fit_transform(data)


class PCAProjector(BaseProjector):
    """Principal Component Analysis projector.
    
    Args:
        Reference:
        https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
    """
    def __init__(self,
                 whiten: bool = False,
                 svd_solver: str = "auto",
                 tol: float = 0.0,
                 iterated_power: Union[str, int] = "auto",
                 random_state: Optional[int] = None) -> None:
        self._projector = PCA(n_components=2,
                              whiten=whiten,
                              svd_solver=svd_solver,
                              tol=tol,
                              iterated_power=iterated_power,
                              random_state=random_state)

    def __call__(self, data: np.ndarray) -> np.ndarray:
        return self._projector.fit_transform(data)


class TSNEProjector(BaseProjector):
    """t-Distributed Stochastic Neighbor Embedding projector.
    
    Args:
        Reference:
        https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html
    """
    def __init__(self,
                 perplexity: float = 30.0,
                 early_exaggeration: float = 12.0,
                 learning_rate: float = 200.0,
                 n_iter: int = 1000,
                 n_iter_without_progress: int = 300,
                 min_grad_norm: float = 1e-7,
                 metric: Union[str, Callable] = "euclidean",
                 init: Union[str, np.ndarray] = "random",
                 verbose: int = 0,
                 random_state: Optional[int] = None,
                 method: str = "barnes_hut",
                 angle: float = 0.5) -> None:
        self._projector = TSNE(n_components=2,
                               perplexity=perplexity,
                               early_exaggeration=early_exaggeration,
                               learning_rate=learning_rate,
                               n_iter=n_iter,
                               n_iter_without_progress=n_iter_without_progress,
                               min_grad_norm=min_grad_norm,
                               metric=metric,
                               init=init,
                               verbose=verbose,
                               random_state=random_state,
                               method=method,
                               angle=angle)

    def __call__(self, data: np.ndarray) -> np.ndarray:
        return self._projector.fit_transform(data)



from datetime import datetime
from mobo.configuration import GlobalConfiguration
import numpy as np
import pandas as pd
from scipy.stats import gaussian_kde
from typing import List, Optional


class Optimizer(object):
    """Optimization pipeline initialized from a configuration.

    Args:
        configuration: Configuration object to initialize from.
    """
    def __init__(self, configuration: GlobalConfiguration) -> None:
        self.configuration = configuration

    def __call__(self) -> None:
        self._log("Beginning the optimization process...")
        # initialize the dataframe
        path = self.configuration.initial_data_path
        if path is None:
            self._log("Generating initial parameter distributions...")
            index = range(self.configuration.n_samples)
            df = pd.DataFrame(columns=self.df_column_headers, index=index)
            init_dist = self._generate_initial_parameter_distributions()
            df[self.parameter_headers] = init_dist
        else:
            self._log("Reading initial parameter distributions from file...")
            df = pd.read_csv(path)
        # loop over each iteration
        last_df: Optional[pd.DataFrame] = None
        for i in range(len(self.configuration.local_configurations)):
            iteration_start = datetime.now()
            self._log("\nBeginning iteration {}...".format(i))
            # do not resample the first iteration
            if i == 0:
                self._log("Skipped resampling step for the first iteration.")
            else:
                df = self._sample(last_df, i) # returns a new df
            # evaluate the parameterizations
            df = self._evaluate(df, i)
            # filter out poor parameterization
            df = pd.concat([df, last_df]) # rejoin the old dataframe
            df = self._filter(df, i)
            # reset index for proper joining
            df.reset_index(inplace=True, drop=True)
            # project the filtered parameters onto a 2D space
            df = self._project(df, i)
            # cluster the projected parameter space
            df = self._cluster(df, i)
            # write the iteration data to file
            df = self._export(df, i)
            # store for sampling
            last_df = df
            iteration_end = datetime.now()
            time_delta = round(
                (iteration_end - iteration_start).total_seconds(), 3
            )
            self._log(
                "Completed iteration {} in {} seconds.\n".format(i, time_delta)
            )

    @property
    def parameter_headers(self) -> List[str]:
        return [p.name for p in self.configuration.parameters]

    @property
    def qoi_headers(self) -> List[str]:
        return [q.name for q in self.configuration.qois]

    @property
    def error_headers(self) -> List[str]:
        return ["{}_error".format(qn) for qn in self.qoi_headers]

    @property
    def projection_headers(self) -> List[str]:
        return ["projection_0", "projection_1"]

    @property
    def cluster_header(self) -> str:
        return "cluster_id"

    @property
    def df_column_headers(self) -> List[str]:
        return list(
            self.parameter_headers + self.qoi_headers + self.error_headers +
            self.projection_headers + [self.cluster_header]
        )

    def _sample(self, df: pd.DataFrame, iteration: int) -> pd.DataFrame:
        """Resample the filtered distribution."""
        self._log("Resampling parameter space via KDE...")
        cluster_ids = set(df[self.cluster_header].to_numpy())
        n_samples = self.configuration.local_configurations[iteration].n_samples
        n_samples_per_cluster = n_samples // len(cluster_ids)
        self._log(
            "\tDrawing {} samples per cluster...".format(n_samples_per_cluster)
        )
        samples = []
        for cluster_id in cluster_ids:
            self._log("\tCluster {}:".format(cluster_id))
            data = df[df[self.cluster_header] == cluster_id]
            self._log("\t\tsamples: {}".format(len(data)))
            data_arr = data[self.parameter_headers].to_numpy(float)
            # linalg error when num samples is less than num parameters
            try:
                kde = gaussian_kde(data_arr.T)
                bandwidth = kde.factor
            except (np.linalg.LinAlgError, ValueError, RuntimeWarning):
                msg = (
                    "\n\t[WARNING] The number of samples in cluster {} is too small. \n"
                    "\tSamples will not be drawn from this cluster.\n"
                ).format(cluster_id)
                self._log(msg)
                continue
            else:
                self._log("\t\tbandwidth: {:.6}".format(bandwidth))
                samples.append(kde.resample(n_samples_per_cluster).T)
        samples_arr = np.vstack(samples)
        index = range(len(samples_arr))
        new_df = pd.DataFrame(columns=self.df_column_headers, index=index)
        new_df[self.parameter_headers] = samples_arr
        return new_df

    def _evaluate(self, df: pd.DataFrame, iteration: int) -> pd.DataFrame:
        """Evaluate each qoi for each parameterization."""
        self._log("Evaluating parameterizations...")
        err_calc = self.configuration.local_configurations[iteration].error_calculator
        qoi_targets = np.array([
            qoi.target for qoi in self.configuration.qois
        ])
        all_qoi_values: List[np.ndarray] = []
        all_error_values: List[np.ndarray] = []
        for _, parameterization in df[self.parameter_headers].iterrows():
            qoi_values = np.array([
                qoi.evaluator(parameterization.to_dict()) 
                for qoi in self.configuration.qois
            ])
            all_qoi_values.append(qoi_values)
            error_values = err_calc(actual=qoi_values, target=qoi_targets)
            all_error_values.append(error_values)
        qoi_arr = np.array(all_qoi_values)
        error_arr = np.array(all_error_values)
        df.update(
            {qh: qoi_arr[:, i] for i, qh in enumerate(self.qoi_headers)}
        )
        df.update(
            {eh: error_arr[:, i] for i, eh in enumerate(self.error_headers)}
        )
        return df

    def _filter(self, df: pd.DataFrame, iteration: int) -> pd.DataFrame:
        """Filter out poor parameterizations."""
        self._log("Filtering parameterizations...")
        self._log("\tSamples before filtration: {}".format(len(df)))
        filters = self.configuration.local_configurations[iteration].filters
        for f in filters:
            mask = f(df[self.error_headers].to_numpy())
            df = df[mask]
        self._log("\tSamples after filtration: {}".format(len(df)))
        return df

    def _project(self, df: pd.DataFrame, iteration: int) -> pd.DataFrame:
        """Project parameter space down to a 2D space."""
        self._log("Projecting parameter space down to 2D...")
        proj = self.configuration.local_configurations[iteration].projector
        proj_arr = proj(df[self.parameter_headers].to_numpy())
        df.update(
            {pn: proj_arr[:, i] for i, pn in enumerate(self.projection_headers)}
        )
        return df

    def _cluster(self, df: pd.DataFrame, iteration: int) -> pd.DataFrame:
        """Assign cluster ids to projected parameters."""
        self._log("Clustering projected parameter space...")
        clust = self.configuration.local_configurations[iteration].clusterer
        cluster_ids = clust(df[self.projection_headers].to_numpy())
        df.update(
            {self.cluster_header: cluster_ids}
        )
        return df

    def _export(self, df: pd.DataFrame, iteration: int) -> pd.DataFrame:
        """Export the results of an iteration as a csv file."""
        filename = "mobo_iteration_{}.csv".format(iteration)
        df.to_csv(filename)
        self._log("Exported iteration data to {}.".format(filename))
        return df

    def _generate_initial_parameter_distributions(self) -> np.ndarray:
        """Generate a uniform distribution over the parameter bounds."""
        lows = np.array([p.lower_bound for p in self.configuration.parameters])
        highs = np.array([p.upper_bound for p in self.configuration.parameters])
        size = (self.configuration.n_samples, lows.shape[0])
        return np.random.uniform(low=lows, high=highs, size=size)

    def _log(self, msg: str) -> None:
        """Log a message to file or stdout."""
        logger = self.configuration.logger
        if logger is None:
            print(msg)
        else:
            logger.log(msg)



from abc import ABC
import numpy as np
from sklearn.cluster import DBSCAN, KMeans
from typing import Callable, Optional, Union


class BaseClusterer(ABC):
    """Abstract base class for Clusterers."""
    def __call__(self, data: np.ndarray) -> np.ndarray:
        pass


class DbscanClusterer(BaseClusterer):
    """DBSCAN clustering technique.

    Args:
        Reference:
        https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html
    """
    def __init__(self,
                 eps: float = 0.5,
                 min_samples: int = 5,
                 metric: Union[str, Callable] = "euclidean",
                 metric_params: Optional[dict] = None,
                 algorithm: str = "auto",
                 leaf_size: int = 30,
                 p: Optional[float] = None) -> None:
        self._clusterer = DBSCAN(eps=eps,
                                 min_samples=min_samples,
                                 metric=metric,
                                 metric_params=metric_params,
                                 algorithm=algorithm,
                                 leaf_size=leaf_size,
                                 p=p,
                                 n_jobs=None)

    def __call__(self, data: np.ndarray) -> np.ndarray:
        return self._clusterer.fit_predict(data)


class KmeansClusterer(BaseClusterer):
    """KMeans clustering technique.
    
    Args:
        Reference:
        https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html
    """
    def __init__(self,
                 n_clusters: int = 8,
                 init: Union[str, np.ndarray] = "k-means++",
                 n_init: int = 10,
                 max_iter: int = 300,
                 tol: float = 1e-4,
                 precompute_distances: Union[str, bool] = "auto",
                 verbose: int = 0,
                 random_state: Union[int, None] = None,
                 copy_x: bool = True,
                 algorithm: str = "auto") -> None:
        self._clusterer = KMeans(n_clusters=n_clusters,
                                 init=init,
                                 n_init=n_init,
                                 max_iter=max_iter,
                                 tol=tol,
                                 precompute_distances=precompute_distances,
                                 verbose=verbose,
                                 random_state=random_state,
                                 copy_x=copy_x,
                                 algorithm=algorithm,
                                 n_jobs=None)

    def __call__(self, data: np.ndarray) -> np.ndarray:
        return self._clusterer.fit_predict(data)



from mobo.error import AbsoluteErrorCalculator, LogCoshErrorCalculator
from mobo.error import RawErrorCalculator, SquaredErrorCalculator
import numpy as np

A = np.random.normal(size=(1000, 3))
B = np.random.normal(size=(1000, 3))


def test_absolute_error_calculator():
    absolute = AbsoluteErrorCalculator()
    err = absolute(A, B)
    assert np.max(err) >= 0
    assert err.shape == A.shape == B.shape


def test_log_cosh_error_calculator():
    log_cosh = LogCoshErrorCalculator()
    err = log_cosh(A, B)
    assert err.shape == A.shape == B.shape


def test_raw_error_calculator():
    raw = RawErrorCalculator()
    err = raw(A, B)
    assert err.shape == A.shape == B.shape


def test_squared_error_calculator():
    squared = SquaredErrorCalculator()
    err = squared(A, B)
    assert np.max(err) >= 0
    assert err.shape == A.shape == B.shape



from typing import Callable


class QoI(object):
    """A quantity of interest to be evaluated during optimization.
    
    Notes:
        - `evaluator` should expect a dict mapping parameter names to values.

    Args:
        name: Name of the qoi.
        evaluator: Evaluation function. 
        target: Target value of the qoi.
    """
    def __init__(self, 
                 name: str,
                 evaluator: Callable, 
                 target: float) -> None:
        self.name = name
        self.evaluator = evaluator
        self.target = target


from abc import ABC
import numpy as np


class BaseErrorCalculator(ABC):
    """Abstract base class for ErrorCalculators."""
    def __call__(self, actual: np.ndarray, target: np.ndarray) -> np.ndarray:
        pass


class AbsoluteErrorCalculator(BaseErrorCalculator):
    """Calculator of absolute error."""
    def __call__(self, actual: np.ndarray, target: np.ndarray) -> np.ndarray:
        return np.absolute(actual - target)


class LogCoshErrorCalculator(BaseErrorCalculator):
    """Calculator of log cosh error."""
    def __call__(self, actual: np.ndarray, target: np.ndarray) -> np.ndarray:
        return np.log(np.cosh(actual - target))


class RawErrorCalculator(BaseErrorCalculator):
    """Calculator of raw error."""
    def __call__(self, actual: np.ndarray, target: np.ndarray) -> np.ndarray:
        return actual - target


class SquaredErrorCalculator(BaseErrorCalculator):
    """Calculator of squared error."""
    def __call__(self, actual: np.ndarray, target: np.ndarray) -> np.ndarray:
        return (actual - target)**2



from mobo.cluster import DbscanClusterer, KmeansClusterer
import numpy as np

NROWS = 1000
DATA = np.random.normal(size=(NROWS, 3))


def test_dbscan_clusterer():
    dbscan = DbscanClusterer()
    cluster_ids = dbscan(DATA)
    assert cluster_ids.shape == (NROWS, )


def test_kmeans_clusterer():
    kmeans = KmeansClusterer()
    cluster_ids = kmeans(DATA)
    assert cluster_ids.shape == (NROWS, )



import setuptools

with open("README.md", "r") as f:
      long_description = f.read()

setuptools.setup(
      name="mobo",
      version="1.0.0",
      author="Seaton Ullberg",
      author_email="sullberg@ufl.edu",
      description="A rational and extensible algorithm for solving multi-objective optimization problems",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/seatonullberg/mobo",
      packages=setuptools.find_packages(),
      license="BSD 2-Clause License"
)



from mobo.projection import MDSProjector, PCAProjector, TSNEProjector
import numpy as np

NROWS = 100
DATA = np.random.normal(size=(NROWS, 3))


def test_mds_projector():
    mds = MDSProjector()
    projection = mds(DATA)
    assert projection.shape[0] == NROWS


def test_pca_projector():
    pca = PCAProjector()
    projection = pca(DATA)
    assert projection.shape[0] == NROWS


def test_tsne_projector():
    tsne = TSNEProjector()
    projection = tsne(DATA)
    assert projection.shape[0] == NROWS





class Parameter(object):
    """A continuous variable with optional constraints.
    
    Args:
        name: Name of the parameter.
        lower_bound: Minimum value of the parameter.
        upper_bound: Maximum value of the parameter.
    """
    def __init__(self, name: str, lower_bound: float, 
                 upper_bound: float) -> None:
        self.name = name
        if lower_bound >= upper_bound:
            err = "`lower_bound` must be strictly less than `upper_bound`."
            raise ValueError(err)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
